"""
Logging and error handling utilities for the Swagchain API.
"""

import asyncio
import functools
import logging
import re
from concurrent.futures import ProcessPoolExecutor
from time import perf_counter
from typing import (Any, Callable, Coroutine, Generator, Literal, Sequence,
                    TypeAlias, TypeVar, cast)

from aiohttp.web_exceptions import HTTPException
from markdown_it import MarkdownIt
from markdown_it.renderer import RendererHTML
from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import get_lexer_by_name
from pygments.styles import get_style_by_name
from rich.console import Console
from rich.logging import RichHandler
from rich.pretty import install
from rich.traceback import install as ins
from typing_extensions import ParamSpec

EXCEPTIONS = (
    asyncio.TimeoutError,
    ConnectionError,
    ConnectionRefusedError,
    ConnectionResetError,
    TimeoutError,
    UnicodeDecodeError,
    UnicodeEncodeError,
    UnicodeError,
    TypeError,
    ValueError,
    ZeroDivisionError,
    IndexError,
    AttributeError,
    ImportError,
    ModuleNotFoundError,
    NotImplementedError,
    RecursionError,
    OverflowError,
    KeyError,
    Exception,
)

Case: TypeAlias = Literal["snake", "pascal", "camel", "constant", "human"]
T = TypeVar("T")
P = ParamSpec("P")


def async_io(
    func: Callable[P, T]
) -> Callable[P, Coroutine[T, Any, Any]]:
    """
    Decorator to convert an IO bound function to a coroutine by running it in a thread pool.
    """

    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        return await asyncio.to_thread(func, *args, **kwargs)

    return wrapper


def async_cpu(
    func: Callable[P, T]
) -> Callable[P, Coroutine[T, Any, Any]]:
    """
    Decorator to convert a CPU bound function to a coroutine by running it in a process pool.
    """

    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        with ProcessPoolExecutor() as pool:
            try:
                return await asyncio.get_running_loop().run_in_executor(
                    pool, func, *args, **kwargs
                )
            except RuntimeError:
                return await asyncio.get_event_loop().run_in_executor(
                    pool, func, *args, **kwargs
                )

    return wrapper

def setup_logging(name: str) -> logging.Logger:
    """
    Set's up logging using the Rich library for pretty and informative terminal logs.

    Arguments:
    name -- Name for the logger instance. It's best practice to use the name of the module where logger is defined.
    """
    install()
    ins()
    console = Console(record=True, force_terminal=True)
    console_handler = RichHandler(
        console=console,
        show_time=True,
        show_path=True,
        markup=True,
        rich_tracebacks=True,
        tracebacks_show_locals=True,
        tracebacks_extra_lines=2,
        tracebacks_theme="monokai",
        show_level=False,
    )
    console_handler.setFormatter(logging.Formatter("%(message)s"))
    console_handler.setLevel(logging.INFO)
    logging.basicConfig(level=logging.INFO, handlers=[console_handler])
    logger_ = logging.getLogger(name)
    logger_.setLevel(logging.INFO)
    return logger_


logger = setup_logging(__name__)


def process_time(
    func: Callable[P, Coroutine[Any, Any, T] | T]
) -> Callable[P, Coroutine[Any, Any, T]]:
    """
    A decorator to measure the execution time of a coroutine.

    Arguments:
    func -- The coroutine whose execution time is to be measured.
    """

    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        """
        Wrapper function to time the function call.
        """
        start = perf_counter()
        if asyncio.iscoroutinefunction(func):
            result = await func(*args, **kwargs)
        else:
            result = func(*args, **kwargs)
        end = perf_counter()
        logger.info("Time taken to execute %s: %s seconds", wrapper.__name__, end - start)
        return result # type: ignore

    return wrapper


def handle_errors(
    func: Callable[P, Coroutine[Any, Any, T]|T]   
) -> Callable[P, Coroutine[Any, Any, T]]:
    """
    A decorator to handle errors in a coroutine.

    Arguments:
    func -- The coroutine whose errors are to be handled.
    """

    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        """
        Wrapper function to handle errors in the function call.
        """
        try:
            if asyncio.iscoroutinefunction(func):
                return await func(*args, **kwargs)
            else:
                return func(*args, **kwargs) # type: ignore
        except EXCEPTIONS as exc:
            logger.error(exc.__class__.__name__)  # type: ignore
            logger.error(str(exc))
            raise HTTPException(reason=str(exc)) from exc

    return wrapper


def chunker(seq: Sequence[T], size: int) -> Generator[Sequence[T], None, None]:
    """
    A generator function that chunks a sequence into smaller sequences of the given size.

    Arguments:
    seq -- The sequence to be chunked.
    size -- The size of the chunks.
    """
    return (seq[pos : pos + size] for pos in range(0, len(seq), size))


def gen_emptystr() -> str:
    """
    A generator function that returns an empty string.
    """
    return cast(str, None)


def retry(
    tries: int = 3, delay: float = 1, factor: float = 2.0
) -> Callable[
    [Callable[P, Coroutine[Any, Any, T]]], Callable[P, Coroutine[Any, Any, T]]
]:
    """
    A decorator to retry a coroutine if it fails, with exponential backoff.

    Arguments:

    tries -- The number of times to retry the coroutine.
    delay -- The initial delay between retries.
    factor -- The multiplier applied to the delay for each retry.
    """

    def decorator(
        func: Callable[P, Coroutine[Any, Any, T]]
    ) -> Callable[P, Coroutine[Any, Any, T]]:
        @functools.wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            """
            Wrapper function to retry the function call.
            """
            current_delay = delay
            for attempt in range(tries):
                try:
                    return await func(*args, **kwargs)
                except EXCEPTIONS as exc:
                    logger.error("Attempt %s/%s failed", attempt + 1, tries)
                    logger.error(str(exc))
                    if attempt + 1 == tries:
                        raise HTTPException(reason=str(exc)) from exc
                    else:
                        await asyncio.sleep(current_delay)
                        current_delay *= factor
            raise HTTPException(reason="All retries have been exhausted.")  # type: ignore

        return wrapper

    return decorator


class Caser:
    """
    Caser
    -----
    A class to convert text to different cases, pluggable as a jinja2 filter.
    """

    @staticmethod
    def snake_case(text: str):
        """

        Args:

                text (str): The text to convert

        Returns:

                str: The converted text

        Converts a text to snake_case

        """
        return re.sub(r"(?<!^)(?=[A-Z])", "_", text).lower()

    @staticmethod
    def pascal_case(text: str):
        """

        Args:

                text (str): The text to convert

        Returns:

                str: The converted text

        Converts a text to PascalCase

        """

        return "".join(word.capitalize() for word in text.split("_"))

    @staticmethod
    def camel_case(text: str):
        """

        Args:

                text (str): The text to convert

        Returns:

                str: The converted text

        Converts a text to camelCase

        """

        words = text.split("_")
        return words[0].lower() + "".join(word.capitalize() for word in words[1:])

    @staticmethod
    def constant_case(text: str):
        """

        Args:

                text (str): The text to convert

        Returns:

                str: The converted text

        Converts a text to CONSTANT_CASE

        """

        return re.sub(r"(?<!^)(?=[A-Z])", "_", text).upper()

    @staticmethod
    def human_case(text: str):
        """

        Args:

                text (str): The text to convert

        Returns:

                str: The converted text

        Converts a text to Human Case

        """
        return " ".join(word.capitalize() for word in text.split("_"))

    def __call__(self, text: str, case: Case):
        """

        Args:

                text (str): The text to convert

                case (Case): The case to convert to

        Returns:


                str: The converted text

        Converts a text to a case

        """
        return getattr(self, f"{case}_case")(text)


def highlight_code(code: str, lang: str, _) -> str:
    """
    A function to highlight code using pygments.

    Arguments:
    code -- The code to be highlighted.
    lang -- The language of the code.
    """
    lexer = get_lexer_by_name(lang, stripall=True)
    formatter = HtmlFormatter( # type: ignore
        style=get_style_by_name("monokai"), # type: ignore
    ) # type: ignore
    return highlight(code, lexer, formatter) # type: ignore

def render_markdown(text: str) -> str:
    """
    A function to render markdown using markdown-it.

    Arguments:
    text -- The markdown text to be rendered.
    """
    md = MarkdownIt()
    md.options.html = True
    md.options.linkify = True
    md.options.typographer = True
    md.options.breaks = True
    md.options.highlight = highlight_code
    md.renderer = RendererHTML()
    STYLES = """
    @import url('https://cdnjs.cloudflare.com/ajax/libs/github-markdown-css/5.2.0/github-markdown.min.css');

    .markdown-body {
        box-sizing: border-box;
        min-width: 200px;
        max-width: 980px;
        margin: 0 auto;
        padding: 2rem;
        display: flex;
        flex-direction: column;
        align-items: start;
    }


    @media (max-width: 767px) {
        .markdown-body {
            padding: 15px;
        }
    }
    
    ::-webkit-scrollbar-track {
        background: #1d1f21;
    }

    ::-webkit-scrollbar-thumb {
        background: #373b41;
        border-radius: 1rem;
    }

    ::-webkit-scrollbar-thumb:hover {
        background: #c5c8c6;
    }

    ::-webkit-scrollbar-thumb:active {
        background: #373b41;
    }

    ::-webkit-scrollbar-corner {
        background: #1d1f21;
    }


    """
    SCRIPTS = """
    <link rel="stylesheet" href="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.8.0/styles/default.min.css">
<script src="https://cdnjs.cloudflare.com/ajax/libs/highlight.js/11.8.0/highlight.min.js"></script>
<script defer>hljs.highlightAll();</script>
"""
    style_sheet = STYLES

    markup = md.render(text)
    styles = f"<style>{style_sheet}</style>"
    html = """<div class="markdown-body">""" + markup + "</div>"
    return styles + html + SCRIPTS





import inspect
import json
from abc import ABC
from typing import Any, Callable, Coroutine, List, Optional, Set, Type, TypeVar

import openai
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from pydantic import create_model  # pylint: disable=no-name-in-module
from typing_extensions import ParamSpec

from ..lib.utils import Caser

F = TypeVar("F", bound="OpenAIFunction")
T = TypeVar("T")
P = ParamSpec("P")
c = Caser()


class FunctionCall(BaseModel):
	name: str
	data: Any


class OpenAIFunction(BaseModel, ABC):
	"""
	OpenAI Function
	---------------

	OpenAI Function orchestration, the `json_schema` needed for the OpenAI API is automatically generated, all the subclasses of this class are automatically registered and can be used in the `function_call` coroutine
	"""

	class Metadata:
		subclasses: Set[Type["OpenAIFunction"]] = set()

	@classmethod
	def __init_subclass__(cls, **kwargs: Any):
		super().__init_subclass__(**kwargs)
		_schema = cls.schema()
		if cls.__doc__ is None:
			cls.__doc__ = f"```json\n{cls.schema_json(indent=2)}\n```"
		cls.openaischema = {
			"name": cls.__name__,
			"description": cls.__doc__,
			"parameters": {
				"type": "object",
				"properties": {
					k: v for k, v in _schema["properties"].items() if k != "self"
				},
				"required": _schema.get("required", []),
			},
		}
		cls.Metadata.subclasses.add(cls)

	@classmethod
	def register(
		cls: Type["OpenAIFunction"], func: Callable[P, Coroutine[Any, Any, Any]]
	) -> Type["OpenAIFunction"]:
		"""
		Creates an OpenAIFunction from a coroutine and infers it's schema from the function signature
		"""
		kwargs = inspect.signature(func).parameters
		klass: Type["OpenAIFunction"] = create_model(
			c(func.__name__, "pascal"),
			__base__=cls,
			**{k: (v.annotation, ...) for k, v in kwargs.items() if k != "self"},
			__module__=cls.__module__,
			__validators__={},
			__cls_kwargs__={},
		)
		if func.__doc__ is None:
			raise ValueError(f"Function {func.__name__} has no docstring")
		klass.__doc__ = func.__doc__

		async def run_(self: "OpenAIFunction", **kwargs: Any):  # type: ignore
			return await func(self, **kwargs)  # type: ignore

		klass.__init_subclass__()
		setattr(klass, "__call__", run_)
		return klass

	async def __call__(self, **kwargs: Any) -> FunctionCall:
		response = await self.run(**kwargs)
		return FunctionCall(name=self.__class__.__name__, data=response)

	async def run(self, **kwargs: Any) -> Any:
		"""Main function to be implemented by subclasses"""
		raise NotImplementedError

async def parse_openai_function(
	response: dict[str, Any],
	functions: List[Type[F]] = OpenAIFunction.Metadata.subclasses,
	**kwargs: Any,
) -> FunctionCall:
	"""Parses the response from the OpenAI API and returns a FunctionCall object, meant to have an uniform interface"""
	choice = response["choices"][0]["message"]
	if "function_call" in choice:
		function_call_ = choice["function_call"]
		name = function_call_["name"]
		arguments = function_call_["arguments"]
		print(name, arguments)
		for i in functions:
			if i.__name__ == name:
				result = await i(**json.loads(arguments))(**kwargs)
				break
		else:
			raise ValueError(f"Function {name} not found")
		return result
	return FunctionCall(name="chat", data=choice["content"])


async def function_call(
	text: str,
	model: str = "gpt-3.5-turbo-16k-0613",
	functions: List[Type[F]] = OpenAIFunction.Metadata.subclasses,
	**kwargs: Any,
) -> FunctionCall:
	"""Given the user input, it will infer the appropiate function to call and it's `json_schema`, then it will call the function and return a FunctionCall object"""
	messages = [
		{"role": "user", "content": text},
		{"role": "system", "content": "You are a function Orchestrator"},
	]
	response = await openai.ChatCompletion.acreate(  # type: ignore
		model=model,
		messages=messages,
		functions=[i.openaischema for i in functions],
	)
	return await parse_openai_function(response, functions=functions, **kwargs)  # type: ignore


async def chat_completion(text: str, context: Optional[str] = None) -> str:
	"""

	Chat Completion
	---------------

	Args:

			text (str): The text to generate from

			context (Optional[str], optional): The context to generate from. Defaults to None.

	Returns:

			str: The generated text

	Simple chat completion outside agent context
	"""

	if context is not None:
		messages = [
			{"role": "user", "content": text},
			{"role": "system", "content": context},
		]
	else:
		messages = [{"role": "user", "content": text}]
	response = await openai.ChatCompletion.acreate(
		model="gpt-3.5-turbo-16k-0613", messages=messages
	)
	return response["choices"][0]["message"]["content"]


async def instruction(
	text: str, temperature: float = 0.2, max_tokens: int = 1024
) -> str:
	"""

	Instruction Completion
	---------------

	Args:

			text (str): The text to generate from

	Returns:

			str: The generated text

	Simple instruction completion outside agent context
	"""
	response = await openai.Completion.acreate(  # type: ignore
		model="gpt-3.5-turbo-instruct",
		prompt=text,
		temperature=temperature,
		max_tokens=max_tokens,
		stream=False,
	)
	return response.choices[0].text  # type: ignore



import asyncio
from typing import Optional

from aiohttp import ClientSession
from boto3 import Session
from pydantic import BaseModel, Field  # pylint: disable=no-name-in-module

from ..lib.utils import setup_logging

logger = setup_logging(__name__)

pool = Session()
sns = pool.client("sns", region_name="us-east-1")

class SnsPayload(BaseModel):
	Type: Optional[str] = Field(default=None, alias="Type")
	MessageId: Optional[str] = Field(default=None, alias="MessageId") 
	Token: Optional[str] = Field(default=None, alias="Token")
	TopicArn: Optional[str] = Field(default=None, alias="TopicArn")
	Subject: Optional[str] = Field(default=None, alias="Subject")
	Message: str = Field(..., alias="Message")	
	SubscribeURL: Optional[str] = Field(default=None, alias="SubscribeURL")
	Timestamp: Optional[str] = Field(default=None, alias="Timestamp")
	SignatureVersion: Optional[str] = Field(default=None, alias="SignatureVersion")
	Signature: Optional[str] = Field(default=None, alias="Signature")
	SigningCertURL: Optional[str] = Field(default=None, alias="SigningCertURL")

class SNSPubSub:
	"""PubSub channel to send function call results to the client."""
	async def pub(self, topic_arn:str, data:SnsPayload):
		if data.Type == 'SubscriptionConfirmation':
			subscribe_url = data.SubscribeURL
			if subscribe_url:
				async with ClientSession() as session:
					async with session.get(subscribe_url) as resp:
						if resp.status == 200:
							logger.info("Subscribed to topic %s", subscribe_url)
						else:
							logger.error("Failed to subscribe to topic %s", subscribe_url)
			else:
				logger.error("No subscribe url in message %s", data)
		elif data.Type == 'Notification':
			logger.info("Publishing to topic %s", topic_arn)
			sns.publish(TopicArn=topic_arn, Message=data.Message)
		else:
			logger.error("Unknown message type %s", data.Type)
		return {"message": "ok"}
	
	async def sub(self, url:str, name:str):
		topic = sns.create_topic(Name=name)
		topic_arn = topic["TopicArn"]
		sns.subscribe(TopicArn=topic_arn, Protocol="http", Endpoint=url)

		async def generator():
			while True:
				messages = sns.subscribe(
					TopicArn=topic_arn, Protocol="http", Endpoint=url
				)
				for message in messages.get("Messages", []):
					yield message["Body"]
				await asyncio.sleep(300)

		return generator
		


import os

from boto3 import Session

from .odm import LazyProxy, async_io


class S3Client(LazyProxy[Session]):
	def __init__(self) -> None:
		self.session = self.__load__()
		super().__init__()

	def __load__(self) -> Session:
		return Session()

	@property
	def client(self):
		return self.session.client(service_name="s3", region_name="us-east-1") # type: ignore

	@async_io
	def put_object(self, namespace:str, filename: str, body: bytes):
		key = f"{namespace}/{filename}"
		self.client.put_object(Bucket=os.environ["AWS_S3_BUCKET"], Key=key, Body=body)
		return self.client.generate_presigned_url(
			"get_object", Params={"Bucket": os.environ["AWS_S3_BUCKET"], "Key": key}
		)	
	
	@async_io
	def get_object(self, namespace:str, filename: str):
		key = f"{namespace}/{filename}"
		return self.client.generate_presigned_url(
			"get_object", Params={"Bucket": os.environ["AWS_S3_BUCKET"], "Key": key}
		)
	


from typing import (AsyncGenerator, Dict, List, Literal, Optional, TypeAlias,
                    TypeVar)

import openai  # pylint-disable=E0401
from pydantic import BaseModel, Field  # pylint: disable=no-name-in-module
from typing_extensions import ParamSpec

T = TypeVar("T")
P = ParamSpec("P")

Role: TypeAlias = Literal["assistant", "user", "system", "function"]
Model: TypeAlias = Literal["gpt-4-0613", "gpt-3.5-turbo-16k-0613"]
Size: TypeAlias = Literal["256x256", "512x512", "1024x1024"]
Format: TypeAlias = Literal["url", "base64"]


class Message(BaseModel):
    """
    A message in a chat

    Args:
            role (Role): The role of the message (assistant, user, system, function)
            content (str): The content of the message
    """

    role: Role = Field(..., description="The role of the message")
    content: str = Field(..., description="The content of the message")


class ChatCompletionRequest(BaseModel):
    """

    Chat completion request

    Args:

            model (Model): The model to use for the completion (gpt-4-0613, gpt-3.5-turbo-16k-0613)
            messages (List[Message]): The messages to use for the completion
            temperature (float): The temperature of the completion
            max_tokens (int): The maximum number of tokens to generate
            stream (bool): Whether to stream the completion or not
    """

    model: Model = Field(..., description="The model to use for the completion")
    messages: List[Message] = Field(
        ..., description="The messages to use for the completion"
    )
    temperature: float = Field(
        default=0.5, description="The temperature of the completion"
    )
    max_tokens: int = Field(
        1024, description="The maximum number of tokens to generate"
    )
    stream: bool = Field(False, description="Whether to stream the completion or not")


class ChatCompletionUssage(BaseModel):
    """

    Chat completion usage

    Args:

            prompt_tokens (int): The number of tokens in the prompt

            completion_tokens (int): The number of tokens in the completion

            total_tokens (int): The total number of tokens

    """

    prompt_tokens: int = Field(..., description="The number of tokens in the prompt")
    completion_tokens: int = Field(
        ..., description="The number of tokens in the completion"
    )
    total_tokens: int = Field(..., description="The total number of tokens")


class ChatCompletionChoice(BaseModel):
    """

    Chat completion choice

    Args:

            index (int): The index of the choice

            message (Message): The message of the choice

            finish_reason (str): The reason the choice was finished

    """

    index: int = Field(..., description="The index of the choice")
    message: Message = Field(..., description="The message of the choice")
    finish_reason: str = Field(..., description="The reason the choice was finished")


class ChatCompletionResponse(BaseModel):
    """

    Chat completion response

    Args:

            id (str): The id of the completion

            object (str): The object of the completion

            created (int): The creation time of the completion

            model (Model): The model used for the completion

            choices (List[ChatCompletionChoice]): The choices of the completion

            usage (ChatCompletionUssage): The usage of the completion

            stream (bool): Whether the completion was streamed or not

    """

    id: str = Field(..., description="The id of the completion")
    object: str = Field(..., description="The object of the completion")
    created: int = Field(..., description="The creation time of the completion")
    model: Model = Field(..., description="The model used for the completion")
    choices: List[ChatCompletionChoice] = Field(
        ..., description="The choices of the completion"
    )
    usage: ChatCompletionUssage = Field(..., description="The usage of the completion")
    stream: bool = Field(..., description="Whether the completion was streamed or not")


class VectorResponse(BaseModel):
    """

    Vector response

    Args:

            id (str): The id of the vector

            object (str): The object of the vector

            created (int): The creation time of the vector

            model (Model): The model used for the vector

            data (List[float]): The data of the vector

    """

    text: str = Field(..., description="The text of the completion")
    score: float = Field(..., description="The score of the completion")


class CreateImageResponse(BaseModel):
    """

    Create image response

    Args:

            id (str): The id of the image

            object (str): The object of the image

            created (float): The creation time of the image

            data (List[Dict[Format, str]]): The data of the image

    """

    created: float = Field(...)
    data: List[Dict[Format, str]] = Field(...)


class CreateImageRequest(BaseModel):
    """

    Create image request

    Args:

            prompt (str): The prompt of the image

            n (int): The number of images to generate

            size (Size): The size of the image

            response_format (Format): The format of the response

    """

    prompt: str = Field(...)
    n: int = Field(default=1)
    size: Size = Field(default="1024x1024")
    response_format: Format = Field(default="url")


class LanguageModel(BaseModel):
    """

    ChatGPT: A GPT-3.5+ chatbot

    This class represents ChatGPT Large Language Model behavior and is a wrapper around the OpenAI Chat API.

    Args:

            model (Model): The model to use for the completion (gpt-4-0613, gpt-3.5-turbo-16k-0613)

            temperature (float): The temperature of the completion

            max_tokens (int): The maximum number of tokens to generate

    """

    model: Model = Field(default="gpt-3.5-turbo-16k-0613")
    temperature: float = Field(default=0.2)
    max_tokens: int = Field(default=1024)

    async def chat(self, text: str, context: Optional[str] = None) -> str:
        """

        Chat with the bot with the option to pass a context

        Args:

                text (str): The text to send to the bot

                context (Optional[str], optional): The context of the bot. Defaults to None.

        Returns:

                str: The response of the bot

        """

        if context is not None:
            messages = [
                Message(role="user", content=text).dict(),
                Message(role="system", content=context).dict(),
            ]
        else:
            messages = [Message(role="user", content=text).dict()]

        response = await openai.ChatCompletion.acreate(  # type: ignore
            model=self.model,
            messages=messages,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            stream=False,
        )

        return response.choices[0].message.content  # type: ignore

    async def assist(self, text: str) -> str:
        """
        Assistance with `gpt-3.5-turbo-instruct` model. Similar to `text-davinci` but with more focus on instructions and state of the GPT3.5 model.

        Args:

                text (str): The text to send to the bot

        Returns:

                str: The response of the bot

        """
        response = await openai.Completion.acreate(  # type: ignore
            model="gpt-3.5-turbo-instruct",
            prompt=text,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            stream=False,
        )
        return response.choices[0].text  # type: ignore

    async def stream(
        self, text: str, context: Optional[str] = None
    ) -> AsyncGenerator[str, None]:
        """
        The chat between the bot and the user is streamed through an async generator, suitable for websockets and server-sent events

        Args:

                text (str): The text to send to the bot

                context (Optional[str], optional): The context of the bot. Defaults to None.

        Yields:

                AsyncGenerator[str, None]: The response of the bot

        """

        if context is not None:
            messages = [
                Message(role="user", content=text).dict(),
                Message(role="system", content=context).dict(),
            ]
        else:
            messages = [Message(role="user", content=text).dict()]

        response = await openai.ChatCompletion.acreate(  # type: ignore
            model=self.model,
            messages=messages,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            stream=True,
        )

        async for message in response:  # type: ignore
            data = message["choices"][0]["delta"].get("content")  # type: ignore
            if data is not None:
                yield data
            else:
                break


    async def image(self, text: str, n: int = 1, size: Size = "1024x1024", response_format: Format = "url") ->str:
        """
        Generate images from text

        Args:

                text (str): The text to generate from

                n (int, optional): The number of images to generate. Defaults to 1.

                size (Size, optional): The size of the image. Defaults to "1024x1024".

                response_format (Format, optional): The format of the response. Defaults to "url".

        Returns:

                List[Dict[Format, str]]: The data of the images

        """
        response = await openai.Image.acreate(  # type: ignore
            prompt=text,
            model=self.model,
            n=n,
            size=size,
            response_format=response_format,
        )
        return response['data'][0]['url'] # type: ignore


from typing import List

import openai

from .vector import *


class Memory(VectorClient):
    """

    Represents Agent's memory
    -------------------------

    Pinecone and OpenAI Embeddings provide the tools needed to build a memory for the agent, this class is a wrapper around them and provides a simple interface to perform similarity search and save new embeddings

    """

    @property
    def builder(self):
        """

        Returns:

            QueryBuilder: A query builder object


        QueryBuilder

        ------------

        A query builder object is a pythonic way to perform queries with MongoDB-like syntax, which is indeed the query language used by Pinecone for it's metadata filtering system.

        https://docs.pinecone.io/reference/query

        """
        return QueryBuilder()

    async def encode(self, texts: List[str], namespace: str) -> List[Embedding]:
        """
        Transforms a list of texts into vectorial embedding objects
        An embedding object is a vector with metadata that is a Key-Value dictionary, where the keys are strings and the values can be `str`, `int`, `float`, `bool` or `list[str]`, useful to store structured data about the text for further filtering and querying.

        Args:

            texts (List[str]): The list of texts to encode

            namespace (str): The namespace of the embeddings

        Returns:

            List[Embedding]: The list of embeddings

        """
        response = await openai.Embedding.acreate(  # type: ignore
            input=texts, model="text-embedding-ada-002"
        )
        response = await openai.Embedding.acreate(
            model="text-embedding-ada-002",
            input=texts,
        )
        data = response["data"]
        return [
            Embedding(
                values=embedding["embedding"],
                metadata={"namespace": namespace, "text": text},
            )
            for embedding, text in zip(data, texts)
        ]

    async def save(self, texts: List[str], namespace: str) -> int:
        """

        Saves a list of texts into the memory

        Args:

            texts (List[str]): The list of texts to save

            namespace (str): The namespace of the embeddings

        Returns:

            int: The number of embeddings saved

        """

        embeddings = await self.encode(texts=texts, namespace=namespace)
        return (await self.upsert(embeddings=embeddings)).upsertedCount

    async def search(self, text: str, namespace: str, top_k: int) -> List[str]:
        """


        Performs a similarity search into the memory

        Args:

            text (str): The text to search

            namespace (str): The namespace of the embeddings

            top_k (int): The number of results to return

        Returns:

            List[str]: The list of results

        """

        embeddings = await self.encode(texts=[text], namespace=namespace)
        expr = (self.builder("namespace") == namespace) & (self.builder("text") != text)
        response = await self.query(
            expr=expr.query,
            vector=embeddings[0].values,
            topK=top_k,
        )
        return [match.metadata["text"] for match in response.matches]  # type: ignore



""".. module:: YourModuleName
   :synopsis: A brief description of what this module does.

.. moduleauthor:: Oscar Bahamonde <oscar.bahamonde.dev@gmail.com>"""

from io import BytesIO
from random import choice
from typing import Any, List, Optional

from boto3 import Session
from pydantic import BaseModel, Field  # pylint: disable=no-name-in-module
from typing_extensions import Literal

from ..lib import *

polly = Session().client("polly", region_name="us-east-1")
comprehend = Session().client("comprehend", region_name="us-east-1")
translate = Session().client("translate", region_name="us-east-1")


class Voice(BaseModel):
    VoiceId: str = Field(...)
    LanguageCode: str = Field(...)
    Engine: Literal["standard", "neural"] = Field(...)

   
    @classmethod
    @async_io
    def load(cls) -> List["Voice"]:
        voices = []
        response = polly.describe_voices()
        for voice in response.get("Voices", []):
            voice_obj = cls(
                VoiceId=voice["Id"], # type: ignore
                LanguageCode=voice["LanguageCode"], # type: ignore
                Engine=voice["SupportedEngines"][0], # type: ignore
            )
            voices.append(voice_obj) # type: ignore
        return voices # type: ignore

    @classmethod
    @async_io
    def detect_language(cls, text: str):
        response =comprehend.detect_dominant_language(Text=text)["Languages"]
        return choice(response)["LanguageCode"] # type: ignore

    @classmethod
    async def assign(cls, text: str)->List["Voice"]:
        voices = await cls.load()
        lang = await cls.detect_language(text)
        return [voice for voice in voices if voice.LanguageCode.startswith(lang)]


class Polly(BaseModel):
    Engine: Literal["standard", "neural"] = Field(
        default="standard",
        description="Specifies the engine for Amazon Polly to use when processing input text for speech synthesis.",
    )
    LanguageCode: str = Field(
        default="en-US",
        description="Optional language code for the Synthesize Speech request.",
    )
    LexiconNames: Optional[List[str]] = Field(
        default=None,
        description="List of one or more pronunciation lexicon names you want the service to apply during synthesis.",
    )
    OutputFormat: Literal["json", "mp3", "ogg_vorbis", "pcm"] = Field(
        default="mp3",
        description="The format in which the returned output will be encoded.",
    )
    SampleRate: str = Field(
        default="22050", description="The audio frequency specified in Hz."
    )
    SpeechMarkTypes: Optional[
        List[Literal["sentence", "ssml", "viseme", "word"]]
    ] = Field(
        default=None,
        description="The type of speech marks returned for the input text.",
    )
    Text: str = Field(..., description="Input text to synthesize.")
    TextType: Literal["ssml", "text"] = Field(
        default="text",
        description="Specifies whether the input text is plain text or SSML.",
    )
    VoiceId: str = Field(
        default="Mia", description="Voice ID to use for the synthesis."
    )

    @classmethod
    def load(cls, text: str, language: str = "en-US"):
        return cls(Text=text, LanguageCode=language)

    def synthesize(self):
        return polly.synthesize_speech(**self.dict(exclude_none=True))

    @async_io
    def get_audio(self):
        byte_stream = BytesIO()
        with self.synthesize()["AudioStream"] as stream:
            byte_stream.write(stream.read())
        byte_stream.seek(0)
        return byte_stream

    async def stream_audio(self):
        with self.synthesize()["AudioStream"] as stream:
            while chunk := stream.read(1024):
                yield chunk


class SentimentScore(BaseModel):
    Mixed: float = Field(
        default=None,
        description="The level of confidence that Amazon Comprehend has in the accuracy of its detection of the MIXED sentiment.",
    )
    Negative: float = Field(
        default=None,
        description="The level of confidence that Amazon Comprehend has in the accuracy of its detection of the NEGATIVE sentiment.",
    )
    Neutral: float = Field(
        default=None,
        description="The level of confidence that Amazon Comprehend has in the accuracy of its detection of the NEUTRAL sentiment.",
    )
    Positive: float = Field(
        default=None,
        description="The level of confidence that Amazon Comprehend has in the accuracy of its detection of the POSITIVE sentiment.",
    )
    Sentiment: Literal["POSITIVE", "NEGATIVE", "NEUTRAL", "MIXED"] = Field(
        default=None,
        description="The inferred sentiment that Amazon Comprehend has the highest level of confidence in.",
    )


class Comprehend(BaseModel):
    Text: str = Field(
        ...,
        description="A UTF-8 text string. Each string must contain fewer that 5,000 bytes of UTF-8 encoded characters.",
    )

   
    
    @async_io
    def run(self,**kwargs:Any):
        response = comprehend.detect_sentiment(Text=self.Text, LanguageCode=self.detect_language(self.Text))
        return SentimentScore(
            Sentiment=response["Sentiment"], **response["SentimentScore"]
        )

    @classmethod
    def detect_language(cls, text: str):
        return comprehend.detect_dominant_language(Text=text)["Languages"][
            0
        ]["LanguageCode"] # type: ignore


class Translate(BaseModel):
    SourceLanguageCode: str = Field(
        default="auto",
        description="The language code for the language of the source text. The language must be a language supported by Amazon Translate.",
    )
    TargetLanguageCode: str = Field(
        default="en",
        description="The language code requested for the language of the target text. The language must be a language supported by Amazon Translate.",
    )
    Text: str = Field(
        ...,
        description="The text to translate. The text string can be a maximum of 5,000 bytes long. Depending on your character set, this may be fewer than 5,000 characters.",
    )

    @classmethod
    def load(
        cls,
        text: str,
        source: str = "auto",
        target: str = "en",
    ):
        return cls(Text=text, SourceLanguageCode=source, TargetLanguageCode=target)

    @async_io
    def run(self, **kwargs:Any):
        response = translate.translate_text(**self.dict(exclude_none=True))
        return response["TranslatedText"]



from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from typing import Generic, Iterator, Set, Type, TypeAlias, TypeVar

from aiohttp import ClientSession
from jinja2 import Template
from langchain.agents.agent_toolkits.openapi import planner
from langchain.agents.agent_toolkits.openapi.spec import reduce_openapi_spec
from langchain.chat_models import ChatOpenAI
from langchain.memory.chat_message_histories import DynamoDBChatMessageHistory
from langchain.requests import RequestsWrapper

from .aws import *
from .database import *
from .lib import *
from .tools import *

MessagePromptMapping: TypeAlias = Dict[str,Role]

MESSAGE_TO_PROMPT_MAPPING:MessagePromptMapping = {
    "HumanMessage": "user",
	"SystemMessage": "system",
	"AIMessage": "assistant"
}


class Agent(BaseModel, ABC):
    llm: LanguageModel = Field(default_factory=LanguageModel)
    memory: Memory = Field(default_factory=Memory)
    namespace: str
    topK: int
    template_str: str
    
    def __hash__(self):
        return hash((self.namespace, self.topK))

    def __eq__(self, other:object)->bool:
        if not isinstance(other, type(self)): return False
        return (self.namespace, self.topK) == (other.namespace, other.topK)
    
    @property
    def tools(self)->Set[Type[OpenAIFunction]]:
        """
        Get the tools of the agent
        """
        return OpenAIFunction.Metadata.subclasses




    @abstractmethod
    async def chain(self, text: str, **kwargs:Any) -> str:
        """
        Chain a text
        """
        ...

    
    @retry()
    @handle_errors
    @process_time
    async def assist(self, text: str) -> str:
        """
        Assist with GPT-3.5 instructions
        ----------------------

        Args:

                text (str): The text to assist with

        Returns:

                str: The generated text

        The agent will call the GPT-3.5 instruction completion API
        """

        return await self.llm.assist(text=text)
    
    @property
    def actions(self) -> List[str]:
        """
        Get the actions of the agent
        """
        return [name for name, _ in inspect.getmembers(self.__class__, inspect.isfunction) if not name.startswith("_")]




class Swagchain(Agent):
    """
    Swagchain: A retrieval augmented generation agent
    ----------------------

    This class is a wrapper around the OpenAI Chat API, Pinecone Emedding and DynamoDB Message History. It streamlines the process of creating a retrieval augmented generation agent.
    """

    llm: LanguageModel = Field(default_factory=LanguageModel)
    memory: Memory = Field(default_factory=Memory)
    namespace: str = Field(...)
    topK: int = Field(default=10)
    template_str: str = Field(...)

    @property
    def history(self) -> DynamoDBChatMessageHistory:
        return DynamoDBChatMessageHistory(
            table_name=self.__class__.__name__, session_id=self.namespace
        )
    
    @property
    def tools(self)->Set[Type[OpenAIFunction]]:
        """
        Get the tools of the agent
        """
        return OpenAIFunction.Metadata.subclasses
    
    @handle_errors
    @async_io
    def fetch(self) -> List[Message]:
        """
        Get messages from the history
        ----------------------

        Args:

                prompt (str): The prompt to get messages from

        Returns:

                List[Message]: The messages
        """
        messages = self.history.messages
        return [ Message(role=MESSAGE_TO_PROMPT_MAPPING[message.__class__.__name__], content=message.content) for message in messages]  
    
    @handle_errors
    @async_io
    def add_ai_message(self, text: str) -> None:
        """
        Add an AI message to the history
        ----------------------

        Args:

                text (str): The text to add
        """
        self.history.add_ai_message(text)
    
    @handle_errors
    @async_io
    def add_user_message(self, text: str) -> None:
        """
        Add a user message to the history
        ----------------------

        Args:

                text (str): The text to add
        """
        self.history.add_user_message(text)

    @retry()
    @handle_errors
    @process_time
    async def chain(self, text: str, **kwargs:Any) -> str:
        """
        Retrieval Augmented Generation
        ----------------------

        Args:

                text (str): The text to retrieve from

        Returns:

                str: The generated text

        The agent will find for the KNN of the text into his memory namespace and generate from them
        a response
        """

        knn = await self.memory.search(
            text=text, namespace=self.namespace, top_k=self.topK
        )
        logger.info(knn)
        logger.info(self.namespace)
        if len(knn) == 0:
            response = await self.llm.chat(text=text)
        else:
            context = await Template(self.template_str, enable_async=True).render_async(suggestions=knn, **kwargs)
            response = await self.llm.chat(text=text, context=context)
        await self.memory.save(texts=[text, response], namespace=self.namespace)
        return response


    async def stream_chain(self, text: str) -> AsyncGenerator[str, None]:
        """
        Retrieval Augmented Generation (stream)
        ----------------------

        Args:

                text (str): The text to retrieve from

        Returns:

                AsyncGenerator[str, None]: The generated text

        The agent will find for the KNN of the text into his memory namespace and generate from them
        a response
        """
         
        full_response = ""
        knn = await self.memory.search(
            text=text, namespace=self.namespace, top_k=self.topK
        )
        if len(knn) == 0:
            async for response in self.llm.stream(text=text):
                full_response += response
                yield response
        else:
            context = await Template(self.template_str, enable_async=True).render_async(suggestions=knn)
            logger.info(context)
            async for response in self.llm.stream(text=text, context=context):
                full_response += response
                yield response
        await self.memory.save(texts=[text, full_response], namespace=self.namespace)
        await self.add_user_message(text)  
        await self.add_ai_message(full_response)

    
    @retry()
    @handle_errors
    @process_time
    async def plugin(self, url:str,text:str)->str:
        """
        OpenAPI Plugin
        ----------------------
        The agent retrieves and reduces the openapi spec of the url and creates an agent executor from it, then runs the user's text query through it
        """
        async with ClientSession() as session:
            response = await session.get(url+"/openapi.json")
            spec = await response.json()
            reduced_spec = reduce_openapi_spec(spec)
            agent_executor = planner.create_openapi_agent(api_spec=reduced_spec, requests_wrapper=RequestsWrapper(aiosession=session),llm=ChatOpenAI())
            return await agent_executor.arun(text)

    @retry()
    @handle_errors
    @process_time
    @async_io
    def create_table(self) -> None:
        """
        Create the table
        ----------------------
        The agent creates the table in DynamoDB
        """
        from boto3 import client
        dynamodb = client("dynamodb")
        dynamodb.create_table(
            TableName=self.__class__.__name__,
            KeySchema=[
                {"AttributeName": "SessionId", "KeyType": "HASH"},
            ],
            AttributeDefinitions=[
                {"AttributeName": "SessionId", "AttributeType": "S"},
            ],
            BillingMode="PAY_PER_REQUEST",
        )
        dynamodb.get_waiter("table_exists").wait(TableName=self.__class__.__name__)
             
A = TypeVar("A", bound=Agent)
S = TypeVar("S", bound=Swagchain)

class AgentSwarm(ABC, Generic[A]):
    def __init__(self, agents: Set[A]) -> None:
        self.agents = agents


    def __iter__(self) -> Iterator[A]:
        return iter(self.agents)
    

    def __getitem__(self, name: str) -> A:
        for agent in self.agents:
            if agent.__class__.__name__ == name:
                return agent
        raise KeyError(f"Agent {name} not found")
    
    def __repr__(self) -> str:
        return f"AgentGroup({' '.join([agent.__class__.__name__ for agent in self.agents])})"
    
    def __len__(self) -> int:
        return len(self.agents)
    
    def __contains__(self, name: str) -> bool:
        return name in [agent.__class__.__name__ for agent in self.agents]
    
    def __add__(self, other: A) -> AgentSwarm[A]:
        self.agents.add(other)
        return self
    
    def __sub__(self, other: A) -> AgentSwarm[A]:
        self.agents.remove(other)
        return self
    
    def __eq__(self, other: object) -> bool:
        if isinstance(other, AgentSwarm):
            return self.agents == other.agents # type: ignore
        return False
    
    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)
    
    def __hash__(self) -> int:
        return hash(self.agents)
    
    @property
    @abstractmethod
    def commander(self) -> A:
        """
        Get the commander of the group
        """
        ...

    @abstractmethod
    async def broadcast(self, text: str) -> List[str]:
        """
        Broadcast a text to all the agents
        """
        ...

    @abstractmethod 
    async def ask(self, text: str, agent: str) -> str:
        """
        Ask a specific agent
        """
        ...

    @abstractmethod
    async def tell(self, text: str, agent: Optional[str]=None) -> str:
        """
        Tell a specific agent
        """
        ...


class SwagchainGroup(AgentSwarm[S]):
    async def broadcast(self, text: str) -> List[str]:
        """
        Broadcast a text to all the agents
        """
        return await asyncio.gather(*[agent.chain(text=text) for agent in self.agents])
    
    async def ask(self, text: str, agent: str) -> str:
        """
        Ask a specific agent
        """
        _agent = self[agent]
        return await _agent.chain(text=text)
    
    async def tell(self, text: str, agent: Optional[str]=None) -> str:
        """
        Tell a specific agent
        """
        if agent is None:
            return await self.commander.chain(text=text)
        _agent = self[agent]
        return await _agent.chain(text=text)
    

    @property
    def commander(self) -> S:
        """
        Get the commander of the group
        """
        return sorted(self.agents, key=lambda agent: agent.topK, reverse=True)[0]


import asyncio
from typing import AsyncGenerator, List

from aiohttp import ClientSession, TCPConnector
from bs4 import BeautifulSoup  # pylint: disable=import-error

from ..database import *
from ..language import *
from ..lib.utils import chunker, handle_errors, process_time, setup_logging

logger = setup_logging(__name__)

HEADERS = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
    "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.114 Safari/537.36"
}
BAD_EXT = (
    "png",
    "jpg",
    "jpeg",
    "gif",
    "pdf",
    "doc",
    "docx",
    "ppt",
    "pptx",
    "xls",
    "xlsx",
    "zip",
    "rar",
    "gz",
    "7z",
    "exe",
    "mp3",
    "mp4",
    "avi",
    "mkv",
    "mov",
    "wmv",
    "flv",
    "swf",
)

connector = TCPConnector(limit=1000)

logger = setup_logging(__name__)


@process_time
@handle_errors
async def sitemap_urls(url: str, session: ClientSession) -> List[str]:
    """
    Fetches all urls from a sitemap
    """
    urls: List[str] = []
    if not url.endswith("xml"):
        url = f"{url.rstrip('/')}/sitemap.xml"
    async with session.get(url) as response:
        text = await response.text()
        soup = BeautifulSoup(text, features="xml")
        for loc in soup.findAll("loc"):
            if loc.text.endswith(BAD_EXT):
                continue
            urls.append(loc.text)
            logger.info("Found url: %s", loc.text)
        for nested_sitemap in soup.findAll("sitemap"):
            urls.extend(await sitemap_urls(nested_sitemap.loc.text, session))
    return list(set(urls))


@process_time
@handle_errors
async def fetch_website(url: str, session: ClientSession, max_size: int = 40960) -> str:
    """
    Fetches a single website
    """
    async with session.get(url) as response:
        html = await response.text()
        soup = BeautifulSoup(html, features="lxml")
        data = soup.get_text(strip=True, separator=" ")
        logger.info("Fetched %s", url)
        logger.info("Length: %s", len(data))
        return data[:max_size]


async def sitemap_pipeline(
    url: str,
    namespace: str,
    session: ClientSession,
    memory: Memory = Memory(),
    chunk_size: int = 64,
) -> AsyncGenerator[float, None]:
    """
    Fetches all urls from a sitemap
    """
    urls: List[str] = await sitemap_urls(url, session)
    length: int = len(urls)
    inserted: int = 0
    while urls:
        try:
            chunk = urls[:chunk_size]
            urls = urls[chunk_size:]
            contents: list[str] = await asyncio.gather(
                *[fetch_website(url, session) for url in chunk]
            )

            for chunk in chunker(contents, 128):
                inserted += await memory.save(texts=list(chunk), namespace=namespace)
                progress = inserted / length
                logger.info("Progress: %s", progress)
                yield progress
                if progress >= 1:
                    logger.info("Succesfully inserted %s documents", inserted)
                    break
        except Exception as exc:  # pylint: disable=broad-except
            logger.error(exc)
            continue


@process_time
@handle_errors
async def scrape_urls(url: str, session: ClientSession) -> List[str]:
    """
    Find all children of a website
    """
    async with session.get(url + "/index.html") as response:
        html:str = await response.text()
        soup = BeautifulSoup(html, features="lxml")
        results = soup.find_all("a", href=True)
        logger.info("Found %s results", len(results))
        responses:List[str] = []
        for result in results:
            if result["href"].startswith("http"):
                responses.append(result["href"])
            else:
                responses.append(f"{url.rstrip('/')}/{result['href'].lstrip('/')}")
        return responses


@process_time
@handle_errors
async def scrape_urls_recursive(url: str, session: ClientSession) -> List[str]:
    """
	Find all children of a website recursively
	"""
    urls = await scrape_urls(url, session)
    for _url in urls:
        if _url not in urls:
            urls.extend(await scrape_urls_recursive(_url, session))
    return list(set(urls))


async def scrape_pipeline(
    url: str,
    namespace: str,
    session: ClientSession,
    memory: Memory = Memory(),
    chunk_size: int = 64,
):
    """
    Scrapes a website
    """
    urls = await scrape_urls(url, session)
    logger.info("Found %s urls", len(urls))
    logger.info("Fetching all children")
    length = len(urls)
    inserted = 0
    while urls:
        chunk = urls[:chunk_size]
        urls = urls[chunk_size:]
        try:
            contents = await asyncio.gather(
                *[fetch_website(url, session) for url in chunk]
            )
            # Chunk each page into groups of 128 characters
            for chunk in chunker(contents, 128):
                inserted += await memory.save(texts=list(chunk), namespace=namespace)
                progress = inserted / length
                logger.info("Progress: %s", progress)
                yield progress
                if progress >= 1:
                    logger.info("Succesfully inserted %s documents", inserted)
                    break
        except Exception as exc:  # pylint: disable=broad-except
            logger.error(exc)
            continue


async def website_loader(url: str, namespace: str)-> AsyncGenerator[float, None]:
    """Reads a website and returns a list of strings"""
    async with ClientSession(connector=connector, headers=HEADERS) as session:
        exts = ["sitemap.xml", "index.html", ""]
        for e in exts:
            response = await session.get(url + "/" + e)
            if response.status == 200:
                if e == "sitemap.xml":
                    async for progress in sitemap_pipeline(url, namespace, session):
                        yield progress
                else:
                    async for progress in scrape_pipeline(url, namespace, session):
                        yield progress
                break
        else:
            logger.error("Could not find a sitemap.xml or index.html file")
            raise Exception("Could not find a sitemap.xml or index.html file")


async def website_function(
    url: str, namespace: str) -> int:
    """
    Loads a website into a vector space
    """
    logger.info("Loading urls from %s", url)
    async with ClientSession(connector=connector, headers=HEADERS) as session:
        exts = ["index.html", "sitemap.xml", ""]
        for e in exts:
            response = await session.get(url + "/" + e)
            if response.status == 200:
                if e == "sitemap.xml":
                    urls = await sitemap_urls(url, session)
                else:
                    urls = await scrape_urls_recursive(url, session)
                break   
        else:
            logger.error("Could not find a sitemap.xml or index.html file")
            raise Exception("Could not find a sitemap.xml or index.html file")  
        logger.info("Found %s urls", len(urls))
        logger.info("Fetching all children")
        length = len(urls)
        inserted = 0
        while urls:
            chunk = urls[:64]
            urls = urls[64:]
            try:
                contents = await asyncio.gather(
                    *[fetch_website(url, session) for url in chunk]
                )
                # Chunk each page into groups of 512 characters
                for chunk in chunker(contents, 512):
                    inserted += await Memory().save(texts=list(chunk), namespace=namespace)
                    progress = inserted / length
                    logger.info("Progress: %s", progress)
                    if progress >= 1:
                        logger.info("Succesfully inserted %s documents", inserted)
                        break
            except Exception as exc:
                logger.error(exc)
                continue
    return inserted

class GoogleResult(BaseModel):
    title:str
    url:str

@process_time
@handle_errors
async def google_search(query:str)->List[GoogleResult]:
    """Searches google for a query"""
    async with ClientSession(connector=connector, headers=HEADERS) as session:
        async with session.get(f"https://www.google.com/search?q={query}") as response:
            html = await response.text()
            soup = BeautifulSoup(html, features="lxml")
            results = soup.find_all("div", class_="yuRUbf")
            return [GoogleResult(title=result.find("h3").text, url=result.find("a", href=True)["href"]) for result in results]

async def google_search_pipeline(
    query: str,
    session: ClientSession,
)->List[str]:
    """
    Fetches all urls from a sitemap
    """
    response = await google_search(query)
    urls = [result.url for result in response]
    htmls:List[str] = []
    for url in urls:
        try:
            htmls.append(await fetch_website(url, session))
        except Exception as exc:
            logger.error(exc)
            continue
    return htmls



website_tool = OpenAIFunction.register(website_function)


"""
This module provides utilities for creating HTTP clients using aiohttp and Pydantic.

.. moduleauthor:: Oscar Bahamonde <oscar.bahamonde.dev@gmail.com>
"""

from typing import (Any, AsyncGenerator, Dict, List, Literal, Optional, Type,
                    TypeVar, Union)

from aiohttp import ClientSession, ClientTimeout, TCPConnector, UnixConnector
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from pydantic import Field  # pylint: disable=no-name-in-module
from pydantic import root_validator
from pydantic.main import ModelMetaclass  # pylint: disable=no-name-in-module
from typing_extensions import ParamSpec

Method = Literal["GET", "POST", "PUT", "PATCH", "DELETE", "HEAD", "OPTIONS"]
Json = Union[Dict[str, Any], List[Dict[str, Any]]]
MaybeJson = Optional[Json]


T = TypeVar("T")
P = ParamSpec("P")
M = TypeVar("M", bound=ModelMetaclass)


class SingletonModelMeta(ModelMetaclass):
    """
    Singleton metaclass for Pydantic models.
    ----------------------------------------

    Monkey patches the Pydantic metaclass to ensure that only one instance of a model is created.
    """

    _instances: Dict[Type["SingletonModelMeta"], "SingletonModelMeta"] = {}

    def __call__(
        cls, *args: Any, **kwargs: Any
    ):  # pylint: disable=E0213 # type: ignore
        if cls not in cls._instances:  # type: ignore
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance  # type: ignore
        return cls._instances[cls]  # type: ignore

class Client(BaseModel, metaclass=SingletonModelMeta):  # pylint: disable=E1139
    """
    Client
    ------

    Base Client that is subclassed to create a client for a specific API.

    Features singleton pattern to ensure that only one instance of a client is created allowing session reuse for the same instance configuration.

    :param base_url: The base URL for the API.
    :type base_url: str
    :param headers: The headers to use for the API.
    :type headers: dict[str, str]
    :param limit: The maximum number of connections to allow.
    :type limit: int
    :param timeout: The timeout for the API.
    :type timeout: int
    :param connector_type: The type of connector to use.
    :type connector_type: Literal["tcp", "unix"]

    .. note:: The `connector` and `session` fields are not included in the schema.
    """

    base_url: str = Field(...)
    headers: dict[str, str] = Field(default={"Content-Type": "application/json"})
    limit: int = Field(default=1000, ge=1, le=10000)
    timeout: int = Field(default=10, ge=1, le=100)
    connector_type: Literal["tcp", "unix"] = Field(default="tcp")

    @staticmethod
    def schema_extra(schema: dict[str, Any], _: Type["Client"]) -> None:  # type: ignore
        """
        Removes the `connector` and `session` fields from the schema.
        """
        schema["properties"].pop("connector")
        schema["properties"].pop("session")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.base_url})"

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({self.base_url})"

    def dict(self, *args: Any, **kwargs: Any):
        # Call the superclass's dict method
        """
        Removes the `headers` field from the dictionary representation.
        """
        dct = super().dict(*args, **kwargs)
        # Remove the headers field
        dct.pop("headers", None)
        return dct

    @root_validator(pre=False, skip_on_failure=True)
    @classmethod
    def validate_args(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """
        Validates the arguments passed to the constructor.
        """

        if values["connector_type"] == "tcp":
            values["connector"] = TCPConnector(limit=values["limit"])
        elif values["connector_type"] == "unix":
            values["connector"] = UnixConnector(
                path=values["base_url"], limit=values["limit"]
            )
        values["timeout"] = ClientTimeout(total=values["timeout"])
        values["session"] = ClientSession(
            base_url=values["base_url"],
            headers=values["headers"],
            connector=values["connector"],
            timeout=values["timeout"],
        )
        return values

    def __call__(self) -> ClientSession:
        """
        Returns the `ClientSession` instance.
        """
        return getattr(self, "session")

    async def request(
        self, method: Method, path: str, json: MaybeJson = None, **kwargs: Any
    ):
        """
        Makes an HTTP request.

        :param method: The HTTP method to use.
        :type method: str
        :param path: The path to request.
        :type path: str
        :param kwargs: Additional keyword arguments to pass to the request.
        :type kwargs: Any
        :return: The response data.
        :rtype: Any
        """
        response = await self().request(method, path, json=json, **kwargs)
        if "json" in response.content_type:
            return await response.json()
        if "text" in response.content_type:
            return await response.text()
        else:
            return await response.read()

    async def get(self, path: str, **kwargs: Any) -> Any:
        """
        Makes a GET request.

        :param path: The path to request.
        :type path: str
        :param kwargs: Additional keyword arguments to pass to the request.
        :type kwargs: Any
        :return: The response data.
        :rtype: Any
        """
        return await self.request("GET", path, **kwargs)

    async def post(self, path: str, **kwargs: Any) -> Any:
        """
        Makes a POST request.

        :param path: The path to request.
        :type path: str
        :param kwargs: Additional keyword arguments to pass to the request.
        :type kwargs: Any
        :return: The response data.
        :rtype: Any
        """
        return await self.request("POST", path, **kwargs)

    async def put(self, path: str, **kwargs: Any) -> Any:
        """
        Makes a PUT request.

        :param path: The path to request.
        :type path: str
        :param kwargs: Additional keyword arguments to pass to the request.
        :type kwargs: Any
        :return: The response data.
        :rtype: Any
        """
        return await self.request("PUT", path, **kwargs)

    async def patch(self, path: str, **kwargs: Any) -> Any:
        """
        Makes a PATCH request.

        :param path: The path to request.
        :type path: str
        :param kwargs: Additional keyword arguments to pass to the request.
        :type kwargs: Any
        :return: The response data.
        :rtype: Any
        """
        return await self.request("PATCH", path, **kwargs)

    async def delete(self, path: str, **kwargs: Any) -> Any:
        """
        Makes a DELETE request.

        :param path: The path to request.
        :type path: str
        :param kwargs: Additional keyword arguments to pass to the request.
        :type kwargs: Any
        :return: The response data.
        :rtype: Any
        """
        return await self.request("DELETE", path, **kwargs)

    async def head(self, path: str, **kwargs: Any) -> Any:
        """
        Makes a HEAD request.

        :param path: The path to request.
        :type path: str
        :param kwargs: Additional keyword arguments to pass to the request.
        :type kwargs: Any
        :return: The response data.
        :rtype: Any
        """
        return await self.request("HEAD", path, **kwargs)

    async def options(self, path: str, **kwargs: Any) -> Any:
        """
        Makes an OPTIONS request.

        :param path: The path to request.
        :type path: str
        :param kwargs: Additional keyword arguments to pass to the request.
        :type kwargs: Any
        :return: The response data.
        :rtype: Any
        """
        return await self.request("OPTIONS", path, **kwargs)

    async def stream(
        self, path: str, method: str = "GET", **kwargs: Any
    ) -> AsyncGenerator[bytes, None]:
        """
        Streams a response.

        :param path: The path to request.
        :type path: str
        :param method: The HTTP method to use.
        :type method: str
        :param kwargs: Additional keyword arguments to pass to the request.
        :type kwargs: Any
        :return: The response data.
        :rtype: AsyncGenerator[bytes, None]
        """
        async with self().request(method, path, **kwargs) as response:
            async for chunk in response.content.iter_any():
                yield chunk

    async def text(self, path: str, method: str = "GET", **kwargs: Any) -> str:
        """
        Returns the response as text.

        :param path: The path to request.
        :type path: str
        :param method: The HTTP method to use.
        :type method: str
        :param kwargs: Additional keyword arguments to pass to the request.
        :type kwargs: Any
        :return: The response data.
        :rtype: str
        """
        async with self().request(method, path, **kwargs) as response:
            return await response.text()

    async def blob(self, path: str, method: str = "GET", **kwargs: Any) -> bytes:
        """
        Returns the response as bytes.

        :param path: The path to request.
        :type path: str
        :param method: The HTTP method to use.
        :type method: str
        :param kwargs: Additional keyword arguments to pass to the request.
        :type kwargs: Any
        :return: The response data.
        :rtype: bytes
        """
        async with self().request(method, path, **kwargs) as response:
            return await response.read()



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import (Any, Dict, Generic, Iterable, List, Literal, Optional, Set,
                    Tuple, Type, TypeVar, Union, cast)
from uuid import UUID

from boto3 import Session
from boto3.dynamodb.types import TypeDeserializer, TypeSerializer
from pydantic import BaseConfig  # pylint: disable=no-name-in-module
from pydantic import (BaseModel, Extra,  # pylint: disable=no-name-in-module
                      Field)
from typing_extensions import ParamSpec

from ..lib.utils import async_io
from ..lib.utils import gen_emptystr as emptystring

T = TypeVar("T")
D = TypeVar("D", bound="DynaModel")
P = ParamSpec("P")
Operator = Literal["=", ">", "<", ">=", "<=", "begins_with", "between"]

  # type: ignore

class LazyProxy(Generic[T], ABC):
	def __init__(self) -> None:
		self.__proxied: T | None = None

	def __getattr__(self, attr: str) -> object:
		return getattr(self.__get_proxied__(), attr)

	def __repr__(self) -> str:
		return repr(self.__get_proxied__())

	def __dir__(self) -> Iterable[str]:
		return self.__get_proxied__().__dir__()

	def __get_proxied__(self) -> T:
		proxied = self.__proxied
		if proxied is not None:
			return proxied

		self.__proxied = proxied = self.__load__()
		return proxied

	def __set_proxied__(self, value: T) -> None:
		self.__proxied = value

	def __as_proxied__(self) -> T:
		return cast(T, self)

	@abstractmethod
	def __load__(self) -> T:
		...


class DynamoDB(Generic[D]):
	entities: Set[Type[D]] = set()
	session: Session
	
	def __init__(self, model: Type[D]) -> None:
		self.model = model
		self.session = self.__load__()    
		super().__init__()

	def __load__(self) -> Session:
		return Session()

	@classmethod
	def __class_getitem__(cls, item: Type[D]) -> Type[DynamoDB[D]]:
		cls.entities.add(item)
		return cls

	@property
	def __table_name__(self) -> str:
		ents:List[str] = []
		for ent in self.entities:
			ents.append(ent.__name__)
			ents.sort()
		return "-".join(ents)


	@property
	def serializer(self) -> TypeSerializer:
		return TypeSerializer()

	@property
	def deserializer(self) -> TypeDeserializer:
		return TypeDeserializer()

	@property
	def client(self):
		return self.session.client(service_name="dynamodb")  # type: ignore

	@property
	def resource(self):
		return self.session.resource(service_name="dynamodb")  # type: ignore


	def serialize(self: DynamoDB[D], instance: D) -> Dict[str, Any]:
		return self.serializer.serialize(instance.to_dict())["M"]  # type: ignore

	def deserialize(self: DynamoDB[D], data: Dict[str, Any]) -> D:
		return self.deserializer.deserialize({"M": data})  # type: ignore
	
	@async_io
	@staticmethod
	def create_table(name:str) -> None:
		dynamodb = Session().client(service_name="dynamodb")
		dynamodb.create_table(
			TableName=name,
			AttributeDefinitions=[
				{"AttributeName": "pk", "AttributeType": "S"},
				{"AttributeName": "sk", "AttributeType": "S"},
			],
			KeySchema=[
				{"AttributeName": "pk", "KeyType": "HASH"},
				{"AttributeName": "sk", "KeyType": "RANGE"},
			],
			BillingMode="PAY_PER_REQUEST",
		)
		dynamodb.get_waiter("table_exists").wait(TableName=name)

	@async_io
	def get(self: DynamoDB[D], pk: str, sk: str) -> D:
		response = self.client.get_item(
			TableName=self.__table_name__,
			Key={
				"pk": self.serializer.serialize(pk),
				"sk": self.serializer.serialize(sk),
			},
		)
		return self.deserialize(response["Item"])

	@async_io
	def put(self: DynamoDB[D], instance: D) -> None:
		print(self.__table_name__)
		self.client.put_item(
			TableName=self.__table_name__, Item=self.serialize(instance)
		) 

	@async_io
	def delete(self: DynamoDB[D], instance: D) -> None:
		self.client.delete_item(
			TableName=self.__table_name__, Key={"pk": instance.pk, "sk": instance.sk}
		)

	@async_io
	def query(
		self: DynamoDB[D],
		pk: str,
		sk: Optional[Union[str, Tuple[str, str]]] = None,
		operator: Operator = "begins_with",
	) -> List[D]:
		key_condition = "pk = :pk"
		expression_values = {":pk": {"S": pk}}

		if sk:
			if operator == "begins_with" and isinstance(sk, str):
				key_condition += " AND begins_with ( sk, :sk )"
				expression_values[":sk"] = {"S": sk}
			elif operator == "between" and isinstance(sk, tuple):
				key_condition += " AND sk BETWEEN :sk0 AND :sk1"
				expression_values[":sk0"] = {"S": sk[0]}
				expression_values[":sk1"] = {"S": sk[1]}
			elif operator in ["=", ">", "<", ">=", "<="] and isinstance(sk, str):
				key_condition += f" AND sk {operator} :sk"
				expression_values[":sk"] = {"S": sk}
			else:
				raise ValueError("Invalid sort key")
		response = self.client.query(
			TableName=self.__table_name__,
			KeyConditionExpression=key_condition,
			ExpressionAttributeValues=expression_values,
		)
		return [self.deserialize(item) for item in response["Items"]]

	@async_io
	def scan(self: DynamoDB[D], **kwargs: Any) -> List[D]:
		response = self.client.scan(TableName=self.__table_name__, **kwargs)
		return [self.deserialize(item) for item in response["Items"]]

	@async_io
	def update(self: DynamoDB[D], pk: str, sk: str, **kwargs: Any) -> None:
		self.client.update_item(
			TableName=self.__table_name__, Key={"pk": pk, "sk": sk}, **kwargs
		)


class DynaModel(BaseModel):
	pk: str = Field(default_factory=emptystring, alias="_pk")
	sk: str = Field(default_factory=emptystring, alias="_sk")

	def __init__(self, **kwargs: Any) -> None:
		super().__init__(**kwargs)
		self.pk = self.pk_
		self.sk = self.sk_

	def to_dict(
		self,
		by_alias: bool = False,
		exclude_unset: bool = False,
		exclude_defaults: bool = False,
		exclude_none: bool = False,
	) -> Dict[str, Any]:
		return super().dict(
			by_alias=by_alias,
			exclude_unset=exclude_unset,
			exclude_defaults=exclude_defaults,
			exclude_none=exclude_none,
		)

	def dict(self, **kwargs: Any):
		exclude_set: set[Any] = kwargs.pop("exclude", set()) or set()
		exclude_set.update({"pk", "sk"})
		return super().dict(exclude=exclude_set, **kwargs)

	def json(self, **kwargs: Any):
		exclude_set: set[Any] = kwargs.pop("exclude", set()) or set()
		exclude_set.update({"pk", "sk"})
		return super().json(exclude=exclude_set, **kwargs)

	class Config(BaseConfig):
		allow_population_by_field_name = True
		json_encoders = {
			datetime: lambda dt: dt.astimezone().isoformat(),
			Decimal: lambda d: float(d),
			UUID: lambda u: str(u),
			Enum: lambda e: e.value,
		}
		use_enum_values = True
		extra = Extra.allow

		@staticmethod
		def schema_extra(schema: Dict[str, Any], model_class: Type[DynaModel]) -> None:  # type: ignore
			schema["properties"].pop("_pk")
			schema["properties"].pop("_sk")

	@classmethod
	def __init_subclass__(cls, **kwargs: Any) -> None:
		super().__init_subclass__(**kwargs)
		cls.db = DynamoDB[cls](cls)

	@property
	def pk_(self) -> str:
		for field in self.__fields__.values():
			if field.field_info.extra.get("pk"):
				key = getattr(self, field.name)
				if isinstance(key, Enum):
					key = key.value
				return self.__class__.__name__ + "#" + str(key)
		raise ValueError("No partition key found")

	@property
	def sk_(self) -> str:
		keys: List[str] = []
		for field in self.__fields__.values():
			if field.field_info.extra.get("sk"):
				key = getattr(self, field.name)
				if isinstance(key, Enum):
					key = key.value
				keys.append(str(key))
		if len(keys) == 0:
			raise ValueError("No sort key found")
		return "#".join(keys)
	

	async def put(self: D) -> D:
		await self.db.put(self)
		return self

	async def delete(self: D) -> D:
		await self.db.delete(self)
		return self

	@classmethod
	async def get(cls: Type[D], pk: str, sk: str) -> D:
		return await cls.db.get(pk, sk)
	
	@classmethod
	async def query(cls: Type[D], pk: str, sk: Optional[Union[str, Tuple[str, str]]] = None, operator: Operator = "begins_with") -> List[D]:
		return await cls.db.query(pk, sk, operator)
	
	@classmethod
	async def scan(cls: Type[D], **kwargs: Any) -> List[D]:
		return await cls.db.scan(**kwargs)  
	
	async def update(self: D, **kwargs: Any) -> D:
		await self.db.update(self.pk, self.sk, **kwargs)
		return self
	
	def __repr__(self) -> str:
		return f"<{self.__class__.__name__} {self.pk} {self.sk}>"
	
	@classmethod
	def __table_name__(cls) -> str:
		ents:List[str] = []
		for ent in cls.__subclasses__():
			ents.append(ent.__name__)
			ents.sort() 
		return "-".join(ents)

	
	@classmethod
	@async_io
	def create_table(cls) -> bool:
		try:
			dynamodb = Session().client(service_name="dynamodb")
			dynamodb.create_table(
				TableName=cls.__table_name__(),
				AttributeDefinitions=[
					{"AttributeName": "pk", "AttributeType": "S"},
					{"AttributeName": "sk", "AttributeType": "S"},
				],
				KeySchema=[
					{"AttributeName": "pk", "KeyType": "HASH"},
					{"AttributeName": "sk", "KeyType": "RANGE"},
				],
				BillingMode="PAY_PER_REQUEST",
			)
			dynamodb.get_waiter("table_exists").wait(TableName=cls.__table_name__())
			return True
		except Exception as exc: # pylint: disable=broad-except
			print(exc)
			return False 